@@ -14,8 +14,10 @@
 
 import collections
 import io
+import json
 import os
 import subprocess
+import tempfile
 
 import fixtures
 import mock
@@ -40,19 +42,23 @@ class BaseTestCase(base.BaseTestCase):
 
 
 class TestUndercloud(BaseTestCase):
+    @mock.patch('instack_undercloud.undercloud._handle_upgrade_fact')
     @mock.patch('instack_undercloud.undercloud._configure_logging')
     @mock.patch('instack_undercloud.undercloud._validate_configuration')
     @mock.patch('instack_undercloud.undercloud._run_command')
     @mock.patch('instack_undercloud.undercloud._post_config')
     @mock.patch('instack_undercloud.undercloud._run_orc')
     @mock.patch('instack_undercloud.undercloud._run_yum_update')
+    @mock.patch('instack_undercloud.undercloud._run_yum_clean_all')
     @mock.patch('instack_undercloud.undercloud._run_instack')
     @mock.patch('instack_undercloud.undercloud._generate_environment')
     @mock.patch('instack_undercloud.undercloud._load_config')
     def test_install(self, mock_load_config, mock_generate_environment,
-                     mock_run_yum_update, mock_run_instack, mock_run_orc,
+                     mock_run_instack, mock_run_clean_all,
+                     mock_run_yum_update, mock_run_orc,
                      mock_post_config, mock_run_command,
-                     mock_validate_configuration, mock_configure_logging):
+                     mock_validate_configuration, mock_configure_logging,
+                     mock_upgrade_fact):
         fake_env = mock.MagicMock()
         mock_generate_environment.return_value = fake_env
         undercloud.install('.')
@@ -62,6 +68,63 @@ class TestUndercloud(BaseTestCase):
         mock_run_orc.assert_called_with(fake_env)
         mock_run_command.assert_called_with(
             ['sudo', 'rm', '-f', '/tmp/svc-map-services'], None, 'rm')
+        mock_upgrade_fact.assert_called_with(False)
+
+    @mock.patch('instack_undercloud.undercloud._handle_upgrade_fact')
+    @mock.patch('instack_undercloud.undercloud._configure_logging')
+    @mock.patch('instack_undercloud.undercloud._validate_configuration')
+    @mock.patch('instack_undercloud.undercloud._run_command')
+    @mock.patch('instack_undercloud.undercloud._post_config')
+    @mock.patch('instack_undercloud.undercloud._run_orc')
+    @mock.patch('instack_undercloud.undercloud._run_yum_update')
+    @mock.patch('instack_undercloud.undercloud._run_yum_clean_all')
+    @mock.patch('instack_undercloud.undercloud._run_instack')
+    @mock.patch('instack_undercloud.undercloud._generate_environment')
+    @mock.patch('instack_undercloud.undercloud._load_config')
+    def test_install_upgrade(self, mock_load_config, mock_generate_environment,
+                             mock_run_instack, mock_run_yum_clean_all,
+                             mock_run_yum_update, mock_run_orc,
+                             mock_post_config, mock_run_command,
+                             mock_validate_configuration,
+                             mock_configure_logging, mock_upgrade_fact):
+        fake_env = mock.MagicMock()
+        mock_generate_environment.return_value = fake_env
+        undercloud.install('.', upgrade=True)
+        self.assertTrue(mock_validate_configuration.called)
+        mock_generate_environment.assert_called_with('.')
+        mock_run_instack.assert_called_with(fake_env)
+        mock_run_orc.assert_called_with(fake_env)
+        mock_run_command.assert_called_with(
+            ['sudo', 'rm', '-f', '/tmp/svc-map-services'], None, 'rm')
+        mock_upgrade_fact.assert_called_with(True)
+
+    @mock.patch('instack_undercloud.undercloud._configure_logging')
+    def test_install_exception(self, mock_configure_logging):
+        mock_configure_logging.side_effect = RuntimeError('foo')
+        self.assertRaises(RuntimeError, undercloud.install, '.')
+        log_dict = {'undercloud_operation': "install",
+                    'exception': 'foo',
+                    'log_file': undercloud.PATHS.LOG_FILE
+                    }
+        self.assertIn(undercloud.FAILURE_MESSAGE % log_dict,
+                      self.logger.output)
+
+    @mock.patch('sys.exit')
+    @mock.patch('instack_undercloud.undercloud._configure_logging')
+    def test_install_exception_no_debug(self, mock_configure_logging,
+                                        mock_exit):
+        mock_configure_logging.side_effect = RuntimeError('foo')
+        conf = config_fixture.Config()
+        self.useFixture(conf)
+        conf.config(undercloud_debug=False)
+        undercloud.install('.')
+        log_dict = {'undercloud_operation': "install",
+                    'exception': 'foo',
+                    'log_file': undercloud.PATHS.LOG_FILE
+                    }
+        self.assertIn(undercloud.FAILURE_MESSAGE % log_dict,
+                      self.logger.output)
+        mock_exit.assert_called_with(1)
 
     def test_generate_password(self):
         first = undercloud._generate_password()
@@ -81,13 +144,12 @@ class TestUndercloud(BaseTestCase):
     @mock.patch('instack_undercloud.undercloud._check_memory')
     @mock.patch('instack_undercloud.undercloud._check_sysctl')
     @mock.patch('instack_undercloud.undercloud._validate_network')
-    @mock.patch('instack_undercloud.undercloud._validate_passwords_file')
-    def test_validate_configuration(self, mock_vpf,
-                                    mock_validate_network,
+    @mock.patch('instack_undercloud.undercloud._validate_no_ip_change')
+    def test_validate_configuration(self, mock_vnic, mock_validate_network,
                                     mock_check_memory, mock_check_hostname,
                                     mock_check_sysctl):
         undercloud._validate_configuration()
-        self.assertTrue(mock_vpf.called)
+        self.assertTrue(mock_vnic.called)
         self.assertTrue(mock_validate_network.called)
         self.assertTrue(mock_check_memory.called)
         self.assertTrue(mock_check_hostname.called)
@@ -180,7 +242,7 @@ class TestCheckMemory(BaseTestCase):
     @mock.patch('psutil.virtual_memory')
     def test_sufficient_memory(self, mock_vm):
         mock_vm.return_value = mock.Mock()
-        mock_vm.return_value.total = 4143927296
+        mock_vm.return_value.total = 8589934592
         undercloud._check_memory()
 
     @mock.patch('psutil.virtual_memory')
@@ -202,20 +264,46 @@ class TestCheckSysctl(BaseTestCase):
         undercloud._check_sysctl()
 
 
-@mock.patch('os.path.isfile')
-class TestPasswordsFileExists(BaseTestCase):
+class TestNoIPChange(BaseTestCase):
+    @mock.patch('os.path.isfile', return_value=False)
     def test_new_install(self, mock_isfile):
-        mock_isfile.side_effect = [False]
-        undercloud._validate_passwords_file()
+        undercloud._validate_no_ip_change()
 
-    def test_update_exists(self, mock_isfile):
-        mock_isfile.side_effect = [True, True]
-        undercloud._validate_passwords_file()
+    @mock.patch('instack_undercloud.undercloud.open')
+    @mock.patch('json.loads')
+    @mock.patch('os.path.isfile', return_value=True)
+    def test_update_matches(self, mock_isfile, mock_loads, mock_open):
+        mock_members = [{'name': 'eth0'},
+                        {'name': 'br-ctlplane',
+                         'addresses': [{'ip_netmask': '192.168.24.1/24'}]
+                         }
+                        ]
+        mock_config = {'network_config': mock_members}
+        mock_loads.return_value = mock_config
+        undercloud._validate_no_ip_change()
 
-    def test_update_missing(self, mock_isfile):
-        mock_isfile.side_effect = [True, False]
+    @mock.patch('instack_undercloud.undercloud.open')
+    @mock.patch('json.loads')
+    @mock.patch('os.path.isfile', return_value=True)
+    def test_update_mismatch(self, mock_isfile, mock_loads, mock_open):
+        mock_members = [{'name': 'eth0'},
+                        {'name': 'br-ctlplane',
+                         'addresses': [{'ip_netmask': '192.168.0.1/24'}]
+                         }
+                        ]
+        mock_config = {'network_config': mock_members}
+        mock_loads.return_value = mock_config
         self.assertRaises(validator.FailedValidation,
-                          undercloud._validate_passwords_file)
+                          undercloud._validate_no_ip_change)
+
+    @mock.patch('instack_undercloud.undercloud.open')
+    @mock.patch('json.loads')
+    @mock.patch('os.path.isfile', return_value=True)
+    def test_update_no_network(self, mock_isfile, mock_loads, mock_open):
+        mock_members = [{'name': 'eth0'}]
+        mock_config = {'network_config': mock_members}
+        mock_loads.return_value = mock_config
+        undercloud._validate_no_ip_change()
 
 
 class TestGenerateEnvironment(BaseTestCase):
@@ -297,23 +385,43 @@ class TestGenerateEnvironment(BaseTestCase):
         # Just spot check, we don't want to replicate the entire opt list here
         self.assertEqual(env['INSPECTION_COLLECTORS'],
                          'default,extra-hardware,logs')
-        self.assertEqual('192.0.2.1/24', env['PUBLIC_INTERFACE_IP'])
-        self.assertEqual('192.0.2.1', env['LOCAL_IP'])
+        self.assertEqual('192.168.24.1/24', env['PUBLIC_INTERFACE_IP'])
+        self.assertEqual('192.168.24.1', env['LOCAL_IP'])
+        # The list is generated from a set, so we can't rely on ordering.
+        # Instead make sure that it looks like a valid list by parsing it.
+        drivers = json.loads(env['ENABLED_DRIVERS'])
+        self.assertEqual(sorted(drivers), ['pxe_drac', 'pxe_ilo',
+                                           'pxe_ipmitool', 'pxe_ssh'])
+        self.assertEqual(env['INSPECTION_NODE_NOT_FOUND_HOOK'], '')
+
+    def test_enabled_discovery(self):
+        conf = config_fixture.Config()
+        self.useFixture(conf)
+        conf.config(enable_node_discovery=True,
+                    discovery_default_driver='foobar')
+        env = undercloud._generate_environment('.')
+        # The list is generated from a set, so we can't rely on ordering.
+        # Instead make sure that it looks like a valid list by parsing it.
+        drivers = json.loads(env['ENABLED_DRIVERS'])
+        # Discovery requires enabling the default driver
+        self.assertEqual(sorted(drivers), ['foobar', 'pxe_drac', 'pxe_ilo',
+                                           'pxe_ipmitool', 'pxe_ssh'])
+        self.assertEqual(env['INSPECTION_NODE_NOT_FOUND_HOOK'], 'enroll')
 
     def test_generate_endpoints(self):
         env = undercloud._generate_environment('.')
         endpoint_vars = {k: v for (k, v) in env.items()
                          if k.startswith('UNDERCLOUD_ENDPOINT')}
-        self.assertEqual(39, len(endpoint_vars))
+        self.assertEqual(90, len(endpoint_vars))
         # Spot check one service
-        self.assertEqual('http://192.0.2.1:5000',
+        self.assertEqual('http://192.168.24.1:5000',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_PUBLIC'])
-        self.assertEqual('http://192.0.2.1:5000',
+        self.assertEqual('http://192.168.24.1:5000',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_INTERNAL'])
-        self.assertEqual('http://192.0.2.1:35357',
+        self.assertEqual('http://192.168.24.1:35357',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_ADMIN'])
         # Also check that the tenant id part is preserved
-        self.assertEqual('http://192.0.2.1:8080/v1/AUTH_%(tenant_id)s',
+        self.assertEqual('http://192.168.24.1:8080/v1/AUTH_%(tenant_id)s',
                          env['UNDERCLOUD_ENDPOINT_SWIFT_PUBLIC'])
 
     def test_generate_endpoints_ssl_manual(self):
@@ -322,14 +430,16 @@ class TestGenerateEnvironment(BaseTestCase):
         conf.config(undercloud_service_certificate='test.pem')
         env = undercloud._generate_environment('.')
         # Spot check one service
-        self.assertEqual('https://192.0.2.2:13000',
+        self.assertEqual('https://192.168.24.2:13000',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_PUBLIC'])
-        self.assertEqual('http://192.0.2.1:5000',
+        self.assertEqual('http://192.168.24.3:5000',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_INTERNAL'])
-        self.assertEqual('http://192.0.2.1:35357',
+        self.assertEqual('http://192.168.24.3:35357',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_ADMIN'])
+        self.assertEqual('https://192.168.24.2:443/keystone/v2.0',
+                         env['UNDERCLOUD_ENDPOINT_KEYSTONE_UI_CONFIG_PUBLIC'])
         # Also check that the tenant id part is preserved
-        self.assertEqual('https://192.0.2.2:13808/v1/AUTH_%(tenant_id)s',
+        self.assertEqual('https://192.168.24.2:13808/v1/AUTH_%(tenant_id)s',
                          env['UNDERCLOUD_ENDPOINT_SWIFT_PUBLIC'])
 
     def test_generate_endpoints_ssl_auto(self):
@@ -338,14 +448,14 @@ class TestGenerateEnvironment(BaseTestCase):
         conf.config(generate_service_certificate=True)
         env = undercloud._generate_environment('.')
         # Spot check one service
-        self.assertEqual('https://192.0.2.2:13000',
+        self.assertEqual('https://192.168.24.2:13000',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_PUBLIC'])
-        self.assertEqual('http://192.0.2.1:5000',
+        self.assertEqual('http://192.168.24.3:5000',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_INTERNAL'])
-        self.assertEqual('http://192.0.2.1:35357',
+        self.assertEqual('http://192.168.24.3:35357',
                          env['UNDERCLOUD_ENDPOINT_KEYSTONE_ADMIN'])
         # Also check that the tenant id part is preserved
-        self.assertEqual('https://192.0.2.2:13808/v1/AUTH_%(tenant_id)s',
+        self.assertEqual('https://192.168.24.2:13808/v1/AUTH_%(tenant_id)s',
                          env['UNDERCLOUD_ENDPOINT_SWIFT_PUBLIC'])
 
     def test_absolute_cert_path(self):
@@ -525,7 +635,8 @@ class TestPostConfig(base.BaseTestCase):
                          mock_copy_stackrc, mock_delete, mock_mistral_client,
                          mock_nova_client):
         instack_env = {
-            'UNDERCLOUD_ENDPOINT_MISTRAL_PUBLIC': 'http://192.0.2.1:8989/v2',
+            'UNDERCLOUD_ENDPOINT_MISTRAL_PUBLIC':
+                'http://192.168.24.1:8989/v2',
         }
         mock_get_auth_values.return_value = ('aturing', '3nigma', 'hut8',
                                              'http://bletchley:5000/v2.0')
@@ -534,8 +645,9 @@ class TestPostConfig(base.BaseTestCase):
         mock_instance_mistral = mock.Mock()
         mock_mistral_client.return_value = mock_instance_mistral
         undercloud._post_config(instack_env)
-        mock_nova_client.assert_called_with(2, 'aturing', '3nigma', 'hut8',
-                                            'http://bletchley:5000/v2.0')
+        mock_nova_client.assert_called_with(
+            2, 'aturing', '3nigma', project_name='hut8',
+            auth_url='http://bletchley:5000/v2.0')
         self.assertTrue(mock_copy_stackrc.called)
         mock_configure_ssh_keys.assert_called_with(mock_instance)
         calls = [mock.call(mock_instance, 'baremetal'),
@@ -733,3 +845,84 @@ class TestPostConfig(base.BaseTestCase):
         mock_instance.flavors.list.return_value = mock_flavors
         undercloud._delete_default_flavors(mock_instance)
         mock_instance.flavors.delete.assert_called_once_with('8ar')
+
+
+class TestUpgradeFact(base.BaseTestCase):
+    @mock.patch('instack_undercloud.undercloud._run_command')
+    @mock.patch('os.path.dirname')
+    @mock.patch('os.path.exists')
+    @mock.patch.object(tempfile, 'mkstemp', return_value=(1, '/tmp/file'))
+    def test_upgrade_fact(self, mock_mkstemp, mock_exists,  mock_dirname,
+                          mock_run):
+        fact_path = '/etc/facter/facts.d/undercloud_upgrade.txt'
+        mock_dirname.return_value = '/etc/facter/facts.d'
+        mock_exists.side_effect = [False, True]
+
+        with mock.patch('instack_undercloud.undercloud.open') as mock_open:
+            undercloud._handle_upgrade_fact(True)
+            mock_open.assert_called_with('/tmp/file', 'w')
+
+        run_calls = [
+            mock.call(['sudo', 'mkdir', '-p', '/etc/facter/facts.d']),
+            mock.call(['sudo', 'mv', '/tmp/file', fact_path]),
+            mock.call(['sudo', 'chmod', '0644', fact_path])
+        ]
+        mock_run.assert_has_calls(run_calls)
+        self.assertEqual(mock_run.call_count, 3)
+
+    @mock.patch('instack_undercloud.undercloud._run_command')
+    @mock.patch('os.path.dirname')
+    @mock.patch('os.path.exists')
+    @mock.patch.object(tempfile, 'mkstemp', return_value=(1, '/tmp/file'))
+    def test_upgrade_fact_install(self, mock_mkstemp, mock_exists,
+                                  mock_dirname, mock_run):
+        mock_dirname.return_value = '/etc/facter/facts.d'
+        mock_exists.return_value = False
+
+        with mock.patch('instack_undercloud.undercloud.open') as mock_open:
+            undercloud._handle_upgrade_fact(False)
+            mock_open.assert_not_called()
+
+        mock_run.assert_not_called()
+
+    @mock.patch('instack_undercloud.undercloud._run_command')
+    @mock.patch('os.path.dirname')
+    @mock.patch('os.path.exists')
+    @mock.patch.object(tempfile, 'mkstemp', return_value=(1, '/tmp/file'))
+    def test_upgrade_fact_upgrade_after_install(self, mock_mkstemp,
+                                                mock_exists, mock_dirname,
+                                                mock_run):
+        fact_path = '/etc/facter/facts.d/undercloud_upgrade.txt'
+        mock_dirname.return_value = '/etc/facter/facts.d'
+        mock_exists.return_value = True
+
+        with mock.patch('instack_undercloud.undercloud.open') as open_m:
+            undercloud._handle_upgrade_fact(True)
+            open_m.assert_called_with('/tmp/file', 'w')
+
+        run_calls = [
+            mock.call(['sudo', 'mv', '/tmp/file', fact_path]),
+            mock.call(['sudo', 'chmod', '0644', fact_path])
+        ]
+        mock_run.assert_has_calls(run_calls)
+        self.assertEqual(mock_run.call_count, 2)
+
+
+class TestInstackEnvironment(BaseTestCase):
+    def test_set_allowed_keys(self):
+        env = undercloud.InstackEnvironment()
+        env['HOSTNAME'] = 'localhost1'
+        env['INSPECTION_COLLECTORS'] = 'a,b,c'
+
+    def test_set_unknown_keys(self):
+        env = undercloud.InstackEnvironment()
+
+        def _set():
+            env['CATS_AND_DOGS_PATH'] = '/home'
+
+        self.assertRaisesRegexp(KeyError, 'CATS_AND_DOGS_PATH', _set)
+
+    def test_get_always_allowed(self):
+        env = undercloud.InstackEnvironment()
+        env.get('HOSTNAME')
+        env.get('CATS_AND_DOGS_PATH')
