@@ -36,6 +36,7 @@ from keystoneclient import session
 from keystoneclient import discover
 from mistralclient.api import client as mistralclient
 from mistralclient.api import base as mistralclient_base
+import novaclient as nc
 from novaclient import client as novaclient
 from novaclient import exceptions
 from oslo_config import cfg
@@ -76,7 +77,7 @@ LOG = None
 CONF = cfg.CONF
 COMPLETION_MESSAGE = """
 #############################################################################
-Undercloud install complete.
+Undercloud %(undercloud_operation)s complete.
 
 The file containing this installation's passwords is at
 %(password_path)s.
@@ -88,39 +89,25 @@ secured.
 
 #############################################################################
 """
-CIDR_DEPRECATION_MESSAGE = """
-*****************************************************************************
-The old default CIDR of 192.0.2.0/24 is deprecated due to it being an
-unroutable address range under RFC 5737.  This default will change in the
-Ocata release of OpenStack, so you should stop using the default CIDR and set
-a valid, routable CIDR instead.
+FAILURE_MESSAGE = """
+#############################################################################
+Undercloud %(undercloud_operation)s failed.
 
-Note that if you have already deployed an overcloud with the 192.0.2.0/24
-CIDR, it will not be possible to change it without re-deploying.  If the
-overcloud cannot be re-deployed, you must explicitly set the network values
-in undercloud.conf to ensure continued use of the 192.0.2.0/24 CIDR during
-future upgrades.
-*****************************************************************************
+Reason: %(exception)s
+
+See the previous output for details about what went wrong.  The full install
+log can be found at %(log_file)s.
+
+#############################################################################
 """
-# We need 4 GB, leave a little room for variation in what 4 GB means on
+# We need 8 GB, leave a little room for variation in what 8 GB means on
 # different platforms.
-REQUIRED_MB = 3750
-
-
-# Allow logging of a warning at the end of the deploy if the deprecated cidr
-# is in use.
-deprecated_cidr = False
+REQUIRED_MB = 7680
 
 
 # When adding new options to the lists below, make sure to regenerate the
 # sample config by running "tox -e genconfig" in the project root.
 _opts = [
-    cfg.StrOpt('image_path',
-               default='.',
-               help=('Local file path to the necessary images. The path '
-                     'should be a directory readable by the current user '
-                     'that contains the full set of images.'),
-               ),
     cfg.StrOpt('undercloud_hostname',
                help=('Fully qualified hostname (including domain) to set on '
                      'the Undercloud. If left unset, the '
@@ -130,7 +117,7 @@ _opts = [
                      'will configure all system hostname settings.'),
                ),
     cfg.StrOpt('local_ip',
-               default='192.0.2.1/24',
+               default='192.168.24.1/24',
                help=('IP information for the interface on the Undercloud '
                      'that will be handling the PXE boots and DHCP for '
                      'Overcloud instances.  The IP portion of the value will '
@@ -139,21 +126,27 @@ _opts = [
                      'prefix portion of the value.')
                ),
     cfg.StrOpt('network_gateway',
-               default='192.0.2.1',
+               default='192.168.24.1',
                help=('Network gateway for the Neutron-managed network for '
                      'Overcloud instances. This should match the local_ip '
                      'above when using masquerading.')
                ),
-    cfg.StrOpt('undercloud_public_vip',
-               default='192.0.2.2',
-               help=('Virtual IP address to use for the public endpoints of '
-                     'Undercloud services. Only used with SSL.')
+    cfg.StrOpt('undercloud_public_host',
+               deprecated_name='undercloud_public_vip',
+               default='192.168.24.2',
+               help=('Virtual IP or DNS address to use for the public '
+                     'endpoints of Undercloud services. Only used with SSL.')
                ),
-    cfg.StrOpt('undercloud_admin_vip',
-               default='192.0.2.3',
-               help=('Virtual IP address to use for the admin endpoints of '
-                     'Undercloud services. Only used with SSL.')
+    cfg.StrOpt('undercloud_admin_host',
+               deprecated_name='undercloud_admin_vip',
+               default='192.168.24.3',
+               help=('Virtual IP or DNS address to use for the admin '
+                     'endpoints of Undercloud services. Only used with SSL.')
                ),
+    cfg.ListOpt('undercloud_nameservers',
+                default=[],
+                help=('DNS nameserver(s) to use for the undercloud node.'),
+                ),
     cfg.ListOpt('undercloud_ntp_servers',
                 default=[],
                 help=('List of ntp servers to use.')),
@@ -170,7 +163,7 @@ _opts = [
                       'will be used in place of the value for '
                       'undercloud_service_certificate.  The resulting '
                       'certificate will be written to '
-                      '/etc/pki/tls/certs/undercloud-[undercloud_public_vip].'
+                      '/etc/pki/tls/certs/undercloud-[undercloud_public_host].'
                       'pem.  This certificate is signed by CA selected by the '
                       '"certificate_generation_ca" option.')
                 ),
@@ -202,35 +195,24 @@ _opts = [
                help=('MTU to use for the local_interface.')
                ),
     cfg.StrOpt('network_cidr',
-               sample_default='192.0.2.0/24',
+               default='192.168.24.0/24',
                help=('Network CIDR for the Neutron-managed network for '
                      'Overcloud instances. This should be the subnet used '
-                     'for PXE booting. The current default for this value '
-                     'is 192.0.2.0/24, but this is deprecated due to it being '
-                     'a non-routable CIDR under RFC 5737. The default value '
-                     'for this option will be changed in the Ocata release. '
-                     'A different, valid CIDR should be selected to avoid '
-                     'problems. If an overcloud has already been deployed '
-                     'with the 192.0.2.0/24 CIDR and therefore the CIDR '
-                     'cannot be changed, you must set this option to '
-                     '192.0.2.0/24 explicitly to avoid it changing in future '
-                     'releases, and all other network options related to the '
-                     'CIDR (e.g. local_ip) must also be set to maintain a '
-                     'valid configuration.')
+                     'for PXE booting.')
                ),
     cfg.StrOpt('masquerade_network',
-               default='192.0.2.0/24',
+               default='192.168.24.0/24',
                help=('Network that will be masqueraded for external access, '
                      'if required. This should be the subnet used for PXE '
                      'booting.')
                ),
     cfg.StrOpt('dhcp_start',
-               default='192.0.2.5',
+               default='192.168.24.5',
                help=('Start of DHCP allocation range for PXE and DHCP of '
                      'Overcloud instances.')
                ),
     cfg.StrOpt('dhcp_end',
-               default='192.0.2.24',
+               default='192.168.24.24',
                help=('End of DHCP allocation range for PXE and DHCP of '
                      'Overcloud instances.')
                ),
@@ -259,7 +241,7 @@ _opts = [
                      'listen.  If in doubt, use the default value.')
                ),
     cfg.StrOpt('inspection_iprange',
-               default='192.0.2.100,192.0.2.120',
+               default='192.168.24.100,192.168.24.120',
                deprecated_name='discovery_iprange',
                help=('Temporary IP range that will be given to nodes during '
                      'the inspection process.  Should not overlap with the '
@@ -283,23 +265,36 @@ _opts = [
                 help=('Whether to support introspection of nodes that have '
                       'UEFI-only firmware.')
                 ),
+    cfg.BoolOpt('enable_node_discovery',
+                default=False,
+                help=('Makes ironic-inspector enroll any unknown node that '
+                      'PXE-boots introspection ramdisk in Ironic. By default, '
+                      'the "fake" driver is used for new nodes (it is '
+                      'automatically enabled when this option is set to True).'
+                      ' Set discovery_default_driver to override. '
+                      'Introspection rules can also be used to specify driver '
+                      'information for newly enrolled nodes.')
+                ),
+    cfg.StrOpt('discovery_default_driver',
+               default='pxe_ipmitool',
+               help=('The default driver to use for newly discovered nodes '
+                     '(requires enable_node_discovery set to True). This '
+                     'driver is automatically added to enabled_drivers.')
+               ),
     cfg.BoolOpt('undercloud_debug',
                 default=True,
                 help=('Whether to enable the debug log level for Undercloud '
                       'OpenStack services.')
                 ),
+    cfg.BoolOpt('undercloud_update_packages',
+                default=True,
+                help=('Whether to update packages during the Undercloud '
+                      'install.')
+                ),
     cfg.BoolOpt('enable_tempest',
                 default=True,
                 help=('Whether to install Tempest in the Undercloud.')
                 ),
-    cfg.BoolOpt('enable_mistral',
-                default=True,
-                help=('Whether to install Mistral services in the Undercloud.')
-                ),
-    cfg.BoolOpt('enable_zaqar',
-                default=True,
-                help=('Whether to install Zaqar services in the Undercloud.')
-                ),
     cfg.BoolOpt('enable_telemetry',
                 default=True,
                 help=('Whether to install Telemetry services '
@@ -314,15 +309,34 @@ _opts = [
                 help=('Whether to install requirements to run the TripleO '
                       'validations.')
                 ),
+    cfg.BoolOpt('enable_cinder',
+                default=False,
+                help=('Whether to install the Volume service to be boot '
+                      'overcloud nodes from remote volumes.')),
+    cfg.BoolOpt('enable_legacy_ceilometer_api',
+                default=True,
+                help=('Whether to enable legacy ceilometer api '
+                      'in the Undercloud. '
+                      'Note: Ceilometer API has been deprecated and will be '
+                      'removed in future release. Please consider moving to '
+                      'gnocchi/Aodh/Panko API instead.')
+                ),
+    cfg.BoolOpt('enable_novajoin',
+                default=False,
+                help=('Whether to install novajoin metadata service in '
+                      'the Undercloud.')
+                ),
+    cfg.StrOpt('ipa_otp',
+               default='',
+               help=('One Time Password to register Undercloud node with '
+                     'an IPA server.  '
+                     'Required when enable_novajoin = True.')
+               ),
     cfg.BoolOpt('ipxe_enabled',
                 default=True,
                 help=('Whether to use iPXE for deploy and inspection.'),
                 deprecated_name='ipxe_deploy',
                 ),
-    cfg.BoolOpt('store_events',
-                default=False,
-                help=('Whether to store events in the Undercloud Ceilometer.')
-                ),
     cfg.IntOpt('scheduler_max_attempts',
                default=30, min=1,
                help=('Maximum number of attempts the scheduler will make '
@@ -334,12 +348,15 @@ _opts = [
                 default=False,
                 help=('Whether to clean overcloud nodes (wipe the hard drive) '
                       'between deployments and after the introspection.')),
+    cfg.ListOpt('enabled_drivers',
+                default=['pxe_ipmitool', 'pxe_drac', 'pxe_ilo', 'pxe_ssh'],
+                help=('List of enabled bare metal drivers.')),
 ]
 
 # Passwords, tokens, hashes
 _auth_opts = [
     cfg.StrOpt('undercloud_db_password',
-               help=('Password used for MySQL root user. '
+               help=('Password used for MySQL databases. '
                      'If left unset, one will be automatically generated.')
                ),
     cfg.StrOpt('undercloud_admin_token',
@@ -362,6 +379,10 @@ _auth_opts = [
                help=('Heat service password. '
                      'If left unset, one will be automatically generated.')
                ),
+    cfg.StrOpt('undercloud_heat_cfn_password',
+               help=('Heat cfn service password. '
+                     'If left unset, one will be automatically generated.')
+               ),
     cfg.StrOpt('undercloud_neutron_password',
                help=('Neutron service password. '
                      'If left unset, one will be automatically generated.')
@@ -378,10 +399,18 @@ _auth_opts = [
                help=('Aodh service password. '
                      'If left unset, one will be automatically generated.')
                ),
+    cfg.StrOpt('undercloud_gnocchi_password',
+               help=('Gnocchi service password. '
+                     'If left unset, one will be automatically generated.')
+               ),
     cfg.StrOpt('undercloud_ceilometer_password',
                help=('Ceilometer service password. '
                      'If left unset, one will be automatically generated.')
                ),
+    cfg.StrOpt('undercloud_panko_password',
+               help=('Panko service password. '
+                     'If left unset, one will be automatically generated.')
+               ),
     cfg.StrOpt('undercloud_ceilometer_metering_secret',
                help=('Ceilometer metering secret. '
                      'If left unset, one will be automatically generated.')
@@ -438,6 +467,10 @@ _auth_opts = [
                help=('Horizon secret key. '
                      'If left unset, one will be automatically generated.')
                ),
+    cfg.StrOpt('undercloud_cinder_password',
+               help=('Cinder service password. '
+                     'If left unset, one will be automatically generated.')
+               ),
 ]
 CONF.register_opts(_opts)
 CONF.register_opts(_auth_opts, group='auth')
@@ -468,7 +501,10 @@ def _configure_logging(level, filename):
                         level=level)
     global LOG
     LOG = logging.getLogger(__name__)
-    LOG.addHandler(logging.StreamHandler())
+    handler = logging.StreamHandler()
+    formatter = logging.Formatter(DEFAULT_LOG_FORMAT)
+    handler.setFormatter(formatter)
+    LOG.addHandler(handler)
 
 
 def _load_config():
@@ -576,9 +612,9 @@ def _check_memory():
     mem = psutil.virtual_memory()
     total_mb = mem.total / 1024 / 1024
     if total_mb < REQUIRED_MB:
-        LOG.error('At least 4 GB of memory is required for undercloud '
-                  'installation.  A minimum of 6 GB is recommended. '
-                  'Only detected %d MB' % total_mb)
+        LOG.error('At least %d MB of memory is required for undercloud '
+                  'installation.  A minimum of 8 GB is recommended. '
+                  'Only detected %d MB' % (REQUIRED_MB, total_mb))
         raise RuntimeError('Insufficient memory available')
 
 
@@ -609,43 +645,36 @@ def _validate_network():
         LOG.error('Undercloud configuration validation failed: %s', message)
         raise validator.FailedValidation(message)
 
-    _validate_cidr()
     params = {opt.name: CONF[opt.name] for opt in _opts}
     validator.validate_config(params, error_handler)
 
 
-def _validate_cidr():
-    """Check for default network_cidr
+def _validate_no_ip_change():
+    """Disallow provisioning interface IP changes
 
-    The old default cidr of 192.0.2.0/24 is deprecated due to being unroutable
-    under RFC 5737.  However, since we need to give users notice of the
-    change, we need some logic to warn them of the problem before actually
-    changing the default, which could be a breaking change on upgrades.
-
-    This function handles the sentinel value of None, which indicates that
-    the user has not overridden the default value, and sets an override on the
-    conf opt to match the previous default.  It also sets a global flag so we
-    can warn about the deprecation at the end of the deploy.
+    Changing the provisioning network IP causes a number of issues, so we
+    need to disallow it early in the install before configurations start to
+    be changed.
     """
-    if CONF.network_cidr is None:
-        global deprecated_cidr
-        deprecated_cidr = True
-        CONF.set_override('network_cidr', '192.0.2.0/24')
-
-
-def _validate_passwords_file():
-    """Disallow updates if the passwords file is missing
-
-    If the undercloud was already deployed, the passwords file needs to be
-    present so passwords that can't be changed are persisted.  If the file
-    is missing it will break the undercloud, so we should fail-fast and let
-    the user know about the problem.
-    """
-    if (os.path.isfile(os.path.expanduser('~/stackrc')) and
-            not os.path.isfile(PATHS.PASSWORD_PATH)):
-        message = ('The %s file is missing.  This will cause all service '
-                   'passwords to change and break the existing undercloud. ' %
-                   PATHS.PASSWORD_PATH)
+    os_net_config_file = '/etc/os-net-config/config.json'
+    # Nothing to do if we haven't already installed
+    if not os.path.isfile(
+            os.path.expanduser(os_net_config_file)):
+        return
+    with open(os_net_config_file) as f:
+        network_config = json.loads(f.read())
+    try:
+        ctlplane = [i for i in network_config.get('network_config', [])
+                    if i['name'] == 'br-ctlplane'][0]
+    except IndexError:
+        # Nothing to check if br-ctlplane wasn't configured
+        return
+    existing_ip = ctlplane['addresses'][0]['ip_netmask']
+    if existing_ip != CONF.local_ip:
+        message = ('Changing the local_ip is not allowed.  Existing IP: '
+                   '%s, Configured IP: %s') % (existing_ip,
+                                               CONF.network_cidr)
+        LOG.error(message)
         raise validator.FailedValidation(message)
 
 
@@ -655,11 +684,11 @@ def _validate_configuration():
         _check_memory()
         _check_sysctl()
         _validate_network()
-        _validate_passwords_file()
+        _validate_no_ip_change()
     except RuntimeError as e:
-        LOG.error('ERROR: An error occured during configuration validation, '
+        LOG.error('An error occurred during configuration validation, '
                   'please check your host configuration and try again. '
-                  '{error}'.format(error=e))
+                  'Error message: {error}'.format(error=e))
         sys.exit(1)
 
 
@@ -704,7 +733,8 @@ def _generate_endpoints(instack_env):
 
     if (CONF.undercloud_service_certificate or
             CONF.generate_service_certificate):
-        public_host = CONF.undercloud_public_vip
+        public_host = CONF.undercloud_public_host
+        internal_host = CONF.undercloud_admin_host
         public_proto = 'https'
         zaqar_ws_public_proto = 'wss'
 
@@ -715,6 +745,18 @@ def _generate_endpoints(instack_env):
             '%s://%s:%d/v1/%%(tenant_id)s',
             {'host': public_host, 'port': 8004, 'ssl_port': 13004},
             {'host': internal_host, 'port': 8004}),
+        ('heat-cfn',
+            '%s://%s:%d/v1/%%(tenant_id)s',
+            {'host': public_host, 'port': 8000, 'ssl_port': 13800},
+            {'host': internal_host, 'port': 8000}),
+        ('heat-ui-proxy',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8004, 'ssl_port': 13004},
+            {'host': internal_host, 'port': 8004}),
+        ('heat-ui-config',
+            '%s://%s:%d/heat/v1/%%(tenant_id)s',
+            {'host': public_host, 'port': 3000, 'ssl_port': 443},
+            {'host': internal_host, 'port': 3000}),
         ('neutron',
             '%s://%s:%d',
             {'host': public_host, 'port': 9696, 'ssl_port': 13696},
@@ -727,6 +769,10 @@ def _generate_endpoints(instack_env):
             '%s://%s:%d/v2.1',
             {'host': public_host, 'port': 8774, 'ssl_port': 13774},
             {'host': internal_host, 'port': 8774}),
+        ('placement',
+            '%s://%s:%d/placement',
+            {'host': public_host, 'port': 8778, 'ssl_port': 13778},
+            {'host': internal_host, 'port': 8778}),
         ('ceilometer',
             '%s://%s:%d',
             {'host': public_host, 'port': 8777, 'ssl_port': 13777},
@@ -736,14 +782,31 @@ def _generate_endpoints(instack_env):
             {'host': public_host, 'port': 5000, 'ssl_port': 13000},
             {'host': internal_host, 'port': 5000},
             {'host': internal_host, 'port': 35357}),
+        ('keystone-ui-config',
+            '%s://%s:%d/keystone/v2.0',
+            {'host': public_host, 'port': 3000, 'ssl_port': 443},
+            {'host': internal_host, 'port': 3000},
+            {'host': internal_host, 'port': 35357}),
         ('swift',
             '%s://%s:%d/v1/AUTH_%%(tenant_id)s',
             {'host': public_host, 'port': 8080, 'ssl_port': 13808},
             {'host': internal_host, 'port': 8080}),
+        ('swift-ui-proxy',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8080, 'ssl_port': 13808},
+            {'host': internal_host, 'port': 8080}),
+        ('swift-ui-config',
+            '%s://%s:%d/swift/v1/AUTH_%%(tenant_id)s',
+            {'host': public_host, 'port': 3000, 'ssl_port': 443},
+            {'host': internal_host, 'port': 3000}),
         ('ironic',
             '%s://%s:%d',
             {'host': public_host, 'port': 6385, 'ssl_port': 13385},
             {'host': internal_host, 'port': 6385}),
+        ('ironic-ui-config',
+            '%s://%s:%d/ironic',
+            {'host': public_host, 'port': 3000, 'ssl_port': 443},
+            {'host': internal_host, 'port': 3000}),
         ('ironic_inspector',
             '%s://%s:%d',
             {'host': public_host, 'port': 5050, 'ssl_port': 13050},
@@ -752,14 +815,42 @@ def _generate_endpoints(instack_env):
             '%s://%s:%d',
             {'host': public_host, 'port': 8042, 'ssl_port': 13042},
             {'host': internal_host, 'port': 8042}),
+        ('gnocchi',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8041, 'ssl_port': 13041},
+            {'host': internal_host, 'port': 8041}),
+        ('panko',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8779, 'ssl_port': 13779},
+            {'host': internal_host, 'port': 8779}),
         ('mistral',
             '%s://%s:%d/v2',
             {'host': public_host, 'port': 8989, 'ssl_port': 13989},
             {'host': internal_host, 'port': 8989}),
+        ('mistral-ui-proxy',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8989, 'ssl_port': 13989},
+            {'host': internal_host, 'port': 8989}),
+        ('mistral-ui-config',
+            '%s://%s:%d/mistral/v2',
+            {'host': public_host, 'port': 3000, 'ssl_port': 443},
+            {'host': internal_host, 'port': 3000}),
         ('zaqar',
             '%s://%s:%d',
             {'host': public_host, 'port': 8888, 'ssl_port': 13888},
             {'host': internal_host, 'port': 8888}),
+        ('cinder',
+            '%s://%s:%d/v1/%%(tenant_id)s',
+            {'host': public_host, 'port': 8776, 'ssl_port': 13776},
+            {'host': internal_host, 'port': 8776}),
+        ('cinder_v2',
+            '%s://%s:%d/v2/%%(tenant_id)s',
+            {'host': public_host, 'port': 8776, 'ssl_port': 13776},
+            {'host': internal_host, 'port': 8776}),
+        ('cinder_v3',
+            '%s://%s:%d/v3/%%(tenant_id)s',
+            {'host': public_host, 'port': 8776, 'ssl_port': 13776},
+            {'host': internal_host, 'port': 8776}),
     ]
     for endpoint_data in endpoint_list:
         endpoints.update(
@@ -779,6 +870,24 @@ def _generate_endpoints(instack_env):
         public_proto=zaqar_ws_public_proto,
         internal_proto=zaqar_ws_internal_proto))
 
+    endpoints.update(_get_service_endpoints(
+        'zaqar-ui-proxy',
+        '%s://%s:%d',
+        {'host': public_host, 'port': 9000, 'ssl_port': 443,
+         'zaqar_ws_public_proto': 'ws'},
+        {'host': internal_host, 'port': 9000},
+        public_proto=zaqar_ws_public_proto,
+        internal_proto=zaqar_ws_internal_proto))
+
+    endpoints.update(_get_service_endpoints(
+        'zaqar-ui-config',
+        '%s://%s:%d/zaqar',
+        {'host': public_host, 'port': 3000, 'ssl_port': 443,
+         'zaqar_ws_public_proto': 'wss'},
+        {'host': internal_host, 'port': 3000},
+        public_proto=zaqar_ws_public_proto,
+        internal_proto=zaqar_ws_internal_proto))
+
     # The swift admin endpoint has a different format from the others
     endpoints['UNDERCLOUD_ENDPOINT_SWIFT_ADMIN'] = (
         '%s://%s:%s' % (internal_proto, internal_host, 8080))
@@ -837,6 +946,40 @@ def _member_role_exists(instack_env):
     instack_env['MEMBER_ROLE_EXISTS'] = six.text_type(role_exists)
 
 
+class InstackEnvironment(dict):
+    """An environment to pass to Puppet with some safety checks.
+
+    Keeps lists of variables we add to the operating system environment,
+    and ensures that we don't anything not defined there.
+    """
+
+    INSTACK_KEYS = {'HOSTNAME', 'ELEMENTS_PATH', 'NODE_DIST', 'JSONFILE',
+                    'REG_METHOD', 'REG_HALT_UNREGISTER', 'PUBLIC_INTERFACE_IP'}
+    """The variables instack and/or used elements can read."""
+
+    DYNAMIC_KEYS = {'INSPECTION_COLLECTORS', 'INSPECTION_KERNEL_ARGS',
+                    'INSPECTION_NODE_NOT_FOUND_HOOK',
+                    'TRIPLEO_INSTALL_USER', 'TRIPLEO_UNDERCLOUD_CONF_FILE',
+                    'TRIPLEO_UNDERCLOUD_PASSWORD_FILE', 'MEMBER_ROLE_EXISTS'}
+    """The variables we calculate in _generate_environment call."""
+
+    PUPPET_KEYS = DYNAMIC_KEYS | {opt.name.upper() for _, group in list_opts()
+                                  for opt in group}
+    """Keys we pass for formatting the resulting hieradata."""
+
+    SET_ALLOWED_KEYS = DYNAMIC_KEYS | INSTACK_KEYS | PUPPET_KEYS
+    """Keys which we allow to add/change in this environment."""
+
+    def __init__(self):
+        super(InstackEnvironment, self).__init__(os.environ)
+
+    def __setitem__(self, key, value):
+        if key not in self.SET_ALLOWED_KEYS:
+            raise KeyError('Key %s is not allowed for an InstackEnvironment' %
+                           key)
+        return super(InstackEnvironment, self).__setitem__(key, value)
+
+
 def _generate_environment(instack_root):
     """Generate an environment dict for instack
 
@@ -846,7 +989,7 @@ def _generate_environment(instack_root):
     :param instack_root: The path containing the instack-undercloud elements
         and json files.
     """
-    instack_env = dict(os.environ)
+    instack_env = InstackEnvironment()
     # Rabbit uses HOSTNAME, so we need to make sure it's right
     instack_env['HOSTNAME'] = CONF.undercloud_hostname or socket.gethostname()
 
@@ -917,6 +1060,17 @@ def _generate_environment(instack_root):
 
     instack_env['INSPECTION_KERNEL_ARGS'] = ' '.join(inspection_kernel_args)
 
+    # Ensure correct rendering of the list and uniqueness of the items
+    enabled_drivers = set(CONF.enabled_drivers)
+    if CONF.enable_node_discovery:
+        enabled_drivers.add(CONF.discovery_default_driver)
+        instack_env['INSPECTION_NODE_NOT_FOUND_HOOK'] = 'enroll'
+    else:
+        instack_env['INSPECTION_NODE_NOT_FOUND_HOOK'] = ''
+
+    instack_env['ENABLED_DRIVERS'] = (
+        '[%s]' % ', '.join('"%s"' % drv for drv in set(enabled_drivers)))
+
     instack_env['PUBLIC_INTERFACE_IP'] = instack_env['LOCAL_IP']
     instack_env['LOCAL_IP'] = instack_env['LOCAL_IP'].split('/')[0]
     if instack_env['UNDERCLOUD_SERVICE_CERTIFICATE']:
@@ -931,14 +1085,19 @@ def _generate_environment(instack_root):
     instack_env['TRIPLEO_UNDERCLOUD_CONF_FILE'] = PATHS.CONF_PATH
     instack_env['TRIPLEO_UNDERCLOUD_PASSWORD_FILE'] = PATHS.PASSWORD_PATH
 
+    # Mustache conditional logic requires ENABLE_NOVAJOIN to be undefined
+    # when novajoin is not enabled.
+    if instack_env['ENABLE_NOVAJOIN'].lower() == 'false':
+        del instack_env['ENABLE_NOVAJOIN']
+
     _generate_endpoints(instack_env)
 
     _write_password_file(instack_env)
 
     if CONF.generate_service_certificate:
-        public_vip = CONF.undercloud_public_vip
+        public_host = CONF.undercloud_public_host
         instack_env['UNDERCLOUD_SERVICE_CERTIFICATE'] = (
-            '/etc/pki/tls/certs/undercloud-%s.pem' % public_vip)
+            '/etc/pki/tls/certs/undercloud-%s.pem' % public_host)
 
     _member_role_exists(instack_env)
 
@@ -981,6 +1140,8 @@ def _generate_init_data(instack_env):
             open(_get_template_path('net-config.json.template')).read()
 
     context['HIERADATA_OVERRIDE'] = hiera_entry
+    context['UNDERCLOUD_NAMESERVERS'] = json.dumps(
+        CONF.undercloud_nameservers)
     partials = {'net_config': net_config_json}
     renderer = pystache.Renderer(partials=partials)
     template = _get_template_path('config.json.template')
@@ -988,6 +1149,7 @@ def _generate_init_data(instack_env):
     with open(template) as f:
         config_json = renderer.render(f.read(), context)
 
+    config_json = config_json.replace('&quot;', '"')
     cfn_path = '/var/lib/heat-cfntools/cfn-init-data'
     tmp_json = tempfile.mkstemp()[1]
     with open(tmp_json, 'w') as f:
@@ -1009,6 +1171,13 @@ def _run_instack(instack_env):
     LOG.info('Instack completed successfully')
 
 
+def _run_yum_clean_all(instack_env):
+    args = ['sudo', 'yum', 'clean', 'all']
+    LOG.info('Running yum clean all')
+    _run_live_command(args, instack_env, 'yum-clean-all')
+    LOG.info('yum-clean-all completed successfully')
+
+
 def _run_yum_update(instack_env):
     args = ['sudo', 'yum', 'update', '-y']
     LOG.info('Running yum update')
@@ -1138,7 +1307,7 @@ def _create_mistral_config_environment(instack_env, mistral):
             }))
 
 
-def _create_default_plan(mistral, timeout=180):
+def _create_default_plan(mistral, timeout=360):
     plan_name = 'overcloud'
     queue_name = str(uuid.uuid4())
 
@@ -1192,7 +1361,12 @@ def _post_config_mistral(instack_env, mistral):
 def _post_config(instack_env):
     _copy_stackrc()
     user, password, tenant, auth_url = _get_auth_values()
-    nova = novaclient.Client(2, user, password, tenant, auth_url)
+    # TODO(andreykurilin): remove this check with support of novaclient 6.0.0
+    if nc.__version__[0] == "6":
+        nova = novaclient.Client(2, user, password, tenant, auth_url=auth_url)
+    else:
+        nova = novaclient.Client(2, user, password, auth_url=auth_url,
+                                 project_name=tenant)
 
     _configure_ssh_keys(nova)
     _delete_default_flavors(nova)
@@ -1204,61 +1378,85 @@ def _post_config(instack_env):
     _ensure_flavor(nova, 'block-storage', 'block-storage')
     _ensure_flavor(nova, 'swift-storage', 'swift-storage')
 
-    if CONF.enable_mistral:
-        mistral_url = instack_env['UNDERCLOUD_ENDPOINT_MISTRAL_PUBLIC']
-        mistral = mistralclient.client(
-            mistral_url=mistral_url,
-            username=user,
-            api_key=password,
-            project_name=tenant,
-            auth_url=auth_url)
-        _post_config_mistral(instack_env, mistral)
+    mistral_url = instack_env['UNDERCLOUD_ENDPOINT_MISTRAL_PUBLIC']
+    mistral = mistralclient.client(
+        mistral_url=mistral_url,
+        username=user,
+        api_key=password,
+        project_name=tenant,
+        auth_url=auth_url)
+    _post_config_mistral(instack_env, mistral)
 
 
-def _die_tuskar_die():
-    """Remove tuskar* packages
+def _handle_upgrade_fact(upgrade=False):
+    """Create an upgrade fact for use in puppet
 
-    Make sure to remove tuskar https://bugs.launchpad.net/tripleo/+bug/1691744
-    # openstack-[tuskar, tuskar-ui, tuskar-ui-extras] & python-tuskarclient
+    Since we don't run different puppets depending on if it's an upgrade or
+    not, we need to be able to pass a flag into puppet to let it know if
+    we're doing an upgrade. This is helpful when trying to handle state
+    transitions from an already installed undercloud. This function creates
+    a static fact named undercloud_upgrade only after the install has occurred.
+    When invoked with upgrade=True, the $::undercloud_upgrade fact should
+    be set to true.
+
+    :param upgrade: Boolean indicating if this is an upgrade action or not
     """
-    try:
-        _run_command(['sudo', 'yum', 'remove', '-y', '*tuskar*'])
-    except subprocess.CalledProcessError as e:
-        LOG.error('Error with tuskar removal task %s - continuing', e.output)
+
+    fact_string = 'undercloud_upgrade={}'.format(upgrade)
+    fact_path = '/etc/facter/facts.d/undercloud_upgrade.txt'
+    if not os.path.exists(os.path.dirname(fact_path)) and upgrade:
+        _run_command(['sudo', 'mkdir', '-p', os.path.dirname(fact_path)])
+
+    # We only need to ensure the fact is correct when we've already installed
+    # the undercloud.
+    if os.path.exists(os.path.dirname(fact_path)):
+        tmp_fact = tempfile.mkstemp()[1]
+        with open(tmp_fact, 'w') as f:
+            f.write(fact_string.lower())
+        _run_command(['sudo', 'mv', tmp_fact, fact_path])
+        _run_command(['sudo', 'chmod', '0644', fact_path])
 
 
-def install(instack_root):
+def install(instack_root, upgrade=False):
     """Install the undercloud
 
     :param instack_root: The path containing the instack-undercloud elements
         and json files.
     """
-    _configure_logging(DEFAULT_LOG_LEVEL, PATHS.LOG_FILE)
-    LOG.info('Logging to %s', PATHS.LOG_FILE)
-    _load_config()
-    _clean_os_refresh_config()
-    _validate_configuration()
-    instack_env = _generate_environment(instack_root)
-    _generate_init_data(instack_env)
-    # Even if we backport https://review.openstack.org/#/c/457478/
-    # into stable branches of puppet-ironic, we still need a way
-    # to handle existing deployments.
-    # This task will fix ironic-dbsync.log ownership on existing
-    # deployments during an upgrade. It can be removed after we
-    # release Pike.
+    undercloud_operation = "upgrade" if upgrade else "install"
     try:
-        _run_command(['sudo', '/usr/bin/chown', 'ironic:ironic',
-                      '/var/log/ironic/ironic-dbsync.log'])
-    except subprocess.CalledProcessError:
-        # assume it's a fresh installation, and the file does not exist
-        pass
-    _die_tuskar_die()
-    _run_yum_update(instack_env)
-    _run_instack(instack_env)
-    _run_orc(instack_env)
-    _post_config(instack_env)
-    _run_command(['sudo', 'rm', '-f', '/tmp/svc-map-services'], None, 'rm')
-    LOG.info(COMPLETION_MESSAGE, {'password_path': PATHS.PASSWORD_PATH,
-             'stackrc_path': os.path.expanduser('~/stackrc')})
-    if deprecated_cidr:
-        LOG.warning(CIDR_DEPRECATION_MESSAGE)
+        _configure_logging(DEFAULT_LOG_LEVEL, PATHS.LOG_FILE)
+        LOG.info('Logging to %s', PATHS.LOG_FILE)
+        _load_config()
+        _clean_os_refresh_config()
+        _validate_configuration()
+        instack_env = _generate_environment(instack_root)
+        _generate_init_data(instack_env)
+        # We didn't complete the M->N upgrades correctly with a
+        # `nova-manage db online_data_migrations` command before. This could
+        # cause the post-upgrade db sync to fail. Better be safe than sorry and
+        # run it before package upgrade.
+        if upgrade:
+            _run_command(['sudo', '/usr/bin/nova-manage', 'db',
+                          'online_data_migrations'])
+        if CONF.undercloud_update_packages:
+            _run_yum_clean_all(instack_env)
+            _run_yum_update(instack_env)
+        _handle_upgrade_fact(upgrade)
+        _run_instack(instack_env)
+        _run_orc(instack_env)
+        _post_config(instack_env)
+        _run_command(['sudo', 'rm', '-f', '/tmp/svc-map-services'], None, 'rm')
+    except Exception as e:
+        LOG.error(FAILURE_MESSAGE,
+                  {'undercloud_operation': undercloud_operation,
+                   'exception': six.text_type(e),
+                   'log_file': PATHS.LOG_FILE})
+        if CONF.undercloud_debug:
+            raise
+        sys.exit(1)
+    else:
+        LOG.info(COMPLETION_MESSAGE,
+                 {'undercloud_operation': undercloud_operation,
+                  'password_path': PATHS.PASSWORD_PATH,
+                  'stackrc_path': os.path.expanduser('~/stackrc')})
